using Curses;
using Posix;

public class Canvas {
	public unowned List<Panel>? selected;
	bool centercolumn;
	List<Panel> center;
	List<Panel> panel;
	bool flatpanel;
	bool flatcenter;
	public weak Window win;
	public Panel prompt;
	public Dialog? dialog; // Do it as stack?
	public string promptstr;
	public bool inputprompt;
	public double split = 4;
	public bool running;

	public Canvas () {
		/* curses init */
		win = initscr ();
		start_color ();
		curs_set (0);
		noecho ();
		mousemask (MouseMask.ALL_MOUSE_EVENTS, null);
		win.keypad (true);

		dialog = null;
		running = true;
		flatpanel = false;
		flatcenter = false;
		selected = null;
		panel = new List<Panel> ();
		center = new List<Panel> ();
		centercolumn = true;
		prompt = new_prompt ();
		promptstr = "";

		/* color paletes */
		init_pair (0, Color.WHITE, Color.BLACK);
		init_pair (1, Color.YELLOW, Color.BLACK);
		init_pair (2, Color.WHITE, Color.GREEN);
		init_pair (3, Color.WHITE, Color.RED);
		init_pair (4, Color.WHITE, Color.YELLOW);
		init_pair (5, Color.WHITE, Color.MAGENTA);
	}

	public bool new_panel (string name) {
		switch (name) {
		case "file":
			var str = Dialog.do_entry (this, "Filename:");
			add_panel (new FilePanel (this, str));
			break;
		case "motd":
			var str = Dialog.do_choose (this, "What is your favorite color?",
				(r) => { 
					switch (r) {
					case 0: return "Blue";
					case 1: return "Red";
					case 2: return "Yellow";
					}
					return null;
				});
			if (str != null)
				Dialog.do_message (this, @"The option is $(str)");
			str = Dialog.do_entry (this, "Enter ascii search string");
			if (str != null)
				Dialog.do_message (this, @"The option is $(str)");
			break;
		case "hex":
			add_panel (new HexPanel (this));
			break;
		case "new":
			add_panel (new Panel (this));
			break;
		case "regs":
			add_panel (new RegsPanel (this));
			break;
		case "help":
			add_panel (new HelpPanel (this));
			break;
		default:
			return false;
		}
		return true;
	}

	private Panel new_prompt () {
		var prompt = new Panel (this);
		prompt.border = false;
		return prompt;
	}

	public void add_panel (Panel p) {
		if (centercolumn) {
			center.prepend (p);
			selected = center.first ();
		} else {
			panel.prepend (p);
			selected = panel.first ();
		}
	}

	public void do_refresh () {
		int sidewidth = 0;
		int centerwidth = COLS;
		int centerheight = LINES-1;
		int sideheight = LINES-1;
		int tw = 0;
		int th = 0;

		refresh ();
		unowned Panel? sel = null;
		if (selected != null && selected.data != null)
			sel = selected.data;

		int len = (int) panel.length ();
		if (len > 0) {
			sidewidth = (int) (COLS / split);
			centerwidth -= sidewidth;
			sideheight /= len;
			tw = 1;
		}
		len = (int) center.length ();
		if (len > 0) {
			centerheight /= len;
		}

		/* realign panels in columns */
		int ypos = 0;
		if (flatpanel) {
			unowned List<Panel> foo = panel.first ();
			if (foo != null) {
				var bar = foo.data;
				bar.color = (centercolumn)? 0:1;
				bar.move (LINES-1, sidewidth, ypos, centerwidth); 
				bar.render ();
			}
		} else {
			int off = (int) ((LINES-1)-(sideheight*panel.length ()));
			unowned List<Panel> plast = panel.last ();
			unowned Panel? last = null;
			if (plast != null)
				last = plast.data;
			foreach (var p in panel) {
				th = (p == last)? 0:1;
				p.color = (p==sel)? 1:0;
				// TODO: only move when necesary 
				p.move (sideheight+off+th, sidewidth, ypos, centerwidth); 
				p.render ();
				ypos += sideheight + off;
				off = 0;
			}
		}

		ypos = 0;
		if (flatcenter) {
			unowned List<Panel>? foo = center.first ();
			if (foo != null) {
				var bar = foo.data;
				bar.color = (centercolumn)?1:0;
				bar.move (LINES-1, centerwidth+tw, ypos, 0); 
				bar.render ();
			}
		} else {
			int off = (int) ((LINES-1)-(centerheight*center.length ()));
			unowned List<Panel>? clast = center.last ();
			if (clast != null) {
				var last = clast.data;
				foreach (var p in center) {
					th = (p == last)? 0:1;
					p.color = (p==sel)? 1:0;
					// TODO: only move when necesary 
					p.move (centerheight+off+th, centerwidth+tw, ypos, 0); 
					p.render ();
					ypos += centerheight + off;
					off = 0;
				}
			}
		}

		/* draw bottom prompt */
		prompt.move (1, COLS, LINES-1, 0); 
		prompt.title = ("%c%d/%d%c [0x%08"+uint64.FORMAT+"]> %s").printf (
			(flatcenter)?'[':'=', center.length (),
			panel.length (), (flatpanel)?']':'=', 0LL, promptstr);
		prompt.render ();

		if (dialog != null) {
			dialog.color = 1;
			dialog.align ();
			dialog.render ();
		}
		win.refresh ();
		//refresh ();
	}

	public void next () {
		if (selected != null && selected.next != null)
			selected = selected.next;
	}

	public void prev () {
		if (selected != null && selected.prev != null)
			selected = selected.prev;
	}

	public void toggle_column () {
		centercolumn = !centercolumn;
		if (centercolumn)
			selected = center.first ();
		else selected = panel.first ();
		if (selected == null)
			toggle_column ();
	}

	public unowned Panel? get_selected () {
		unowned Panel? sel = null;
		if (selected != null && selected.data != null)
			sel = selected.data;
		return sel;
	}

	public void toggle_monocle (bool centercolumn) {
		unowned Panel? sel = get_selected ();
		if (centercolumn) {
			center.remove (sel);
			center.prepend (sel);
			selected = center.first ();
			flatcenter = !flatcenter;
		} else {
			panel.remove (sel);
			panel.prepend (sel);
			selected = panel.first ();
			flatpanel = !flatpanel;
		}
	}

	public void do_click (int y, int x, long bstate) {
		unowned Panel? sel = get_selected ();
		if (dialog != null) {
			/* TODO: check for click on window */
			dialog = null;
			return;
		}
		if ((bstate & Button4.CLICKED) != 0) {
			if (sel != null)
				sel.line--;
		} else
/*
		// ONLY FOR CURSES>1 //
		if ((bstate & Button5.CLICKED) != 0) {
			if (sel != null)
				sel.line++;
		} else
*/
		if ((bstate & Button1.DOUBLE_CLICKED) != 0) {
			// TODO: select column too
			toggle_monocle (centercolumn);
		} else
		if ((bstate & Button1.CLICKED) != 0) {
			unowned List<Panel> ptr = center.first ();
			for(; ptr != null; ptr = ptr.next) {
				var p = ptr.data;
				if ((x >= p.x && x <= p.x+p.w)
				&&  (y >= p.y && y <= p.y+p.h)) {
					selected = ptr;
					centercolumn = true;
					return;
				}
				if (flatcenter)
					break;
			}
			ptr = panel.first ();
			for(; ptr != null; ptr = ptr.next) {
				var p = ptr.data;
				if ((x >= p.x && x <= p.x+p.w)
				&&  (y >= p.y && y <= p.y+p.h)) {
					selected = ptr;
					inputprompt = false;
					centercolumn = false;
					break;
				}
				if (flatpanel)
					break;
			}
			if (y == LINES-1) {
				// TODO: doesnt works well :(
				inputprompt = true;
				selected = null;
			}
		}
	}

	public void main_loop () {
		var ch = 0;
		while (running) {
			unowned Panel? sel = get_selected ();
			prompt.color = (inputprompt)? 1:0;

			do_refresh ();
			if (sel != null) {
				sel.win.mvprintw (0, sel.w-5, "(%d)".printf(ch));
				sel.win.refresh ();
			}

			ch = win.getch ();
			if (ch == -1) // refresh
				continue;

			if (sel != null) {
				sel.win.mvprintw (0, sel.w-5, "(%d)".printf(ch));
				sel.win.refresh ();
			}
			/* doesnt works at all */
			if (ch == Key.MOUSE) {
				MouseEvent ev;
				if (getmouse (out ev)) {
					print ("TODO: Do something awesome here\n");
				}
				do_click (ev.y, ev.x, ev.bstate);

				//sel.win.mvprintw (3, 3, "MOUSE %d %d".printf(ev.x, ev.y));
				//sel.win.refresh ();
				continue;
			}

			if (dialog != null) {
				if (ch=='q' || ch==' ' || ch==27 || ch=='\n') {
					dialog = null;
				}
			} else
			if (inputprompt) {
				if (ch==27 || ch=='\n') {
					inputprompt = false;
					unowned List<Panel> tmp = center.first ();
					if (tmp != null)
						selected = tmp;
					new_panel (promptstr);
					promptstr = "";
					continue;
				}
				if (ch == 127 || ch == 263)
					promptstr = promptstr.substring (0, (int)(promptstr.size ()-1));
				else promptstr = "%s%c".printf (promptstr, ch);
			} else
			switch (ch) {
			case ' ':
				if (sel != null)
					sel.activate ();
				break;
			case ConsKey.ControlSpace:
				toggle_monocle (centercolumn);
				break;
			case ConsKey.ControlN:
				var p = new Panel (this);
				this.add_panel (p);
				break;
			case '?': // control+n
				var p = new HelpPanel (this);
				this.add_panel (p);
				break;
			case ']':
				split += 0.25;
				break;
			case '[':
				split -= 0.25;
				if (split<1.1)
					split = 1.1;
				break;
			case ConsKey.ShiftDown:
			case 'J':
				center.remove (sel);
				panel.remove (sel);
				if (centercolumn) {
					center.append (sel);
					selected = center.last ();
				} else {
					panel.append (sel);
					selected = panel.last ();
				}
				break;
			case ConsKey.ShiftUp:
			case 'K':
				center.remove (sel);
				panel.remove (sel);
				if (centercolumn) {
					center.prepend (sel);
					selected = center.first ();
				} else {
					panel.prepend (sel);
					selected = panel.first ();
				}
				break;
			case ConsKey.ControlF:
			case ConsKey.NextPage:
				if (sel != null)
					sel.line += (int)(sel.h / 3);
				break;
			case ConsKey.ControlB:
			case ConsKey.PrevPage:
				if (sel != null)
					sel.line -= (int)(sel.h / 3);
				break;
			case '=':
				if (sel != null)
					sel.line = 0;
				break;
			case ConsKey.Down:
			case 'j':
				if (sel != null)
					sel.line++;
				break;
			case ConsKey.Up:
			case 'k':
				if (sel != null)
					sel.line--;
				break;
#if 0
			case 'h':
				if (center.length () > 0) {
					centercolumn = true;
					selected = center.first ();
				}
				break;
			case 'l':
				if (panel.length () > 0) {
					centercolumn = false;
					selected = center.first ();
				}
				break;
#endif
			//case Key.DOWN:
			case ConsKey.ControlJ:
			case ConsKey.ControlDown:
			case ConsKey.ControlDown1:
			case ConsKey.ControlDown2:
				if (centercolumn) {
					if (flatcenter) {
						center.remove (sel);
						center.append (sel);
						selected = center.first ();
					} else next ();
				} else {
					if (flatpanel) {
						panel.remove (sel);
						panel.append (sel);
						selected = panel.first ();
					} else next ();
				}
				break;
			//case Key.UP:
			case ConsKey.ControlUp:
			case ConsKey.ControlUp1:
			case ConsKey.ControlK:
				if (centercolumn) {
					if (flatcenter) {
						sel = center.last ().data;
						center.remove (sel);
						center.prepend (sel);
						selected = center.first ();
					} else prev ();
				} else {
					if (flatpanel) {
						sel = panel.last ().data;
						panel.remove (sel);
						panel.prepend (sel);
						selected = panel.first ();
					} else prev ();
				}
				break;
			//case Key.LEFT:
			case ConsKey.ControlLeft:
			case ConsKey.ControlLeft1:
			case ConsKey.ControlLeft2:
			case ConsKey.ControlLeft3:
			case ConsKey.ControlH:
			case ConsKey.ControlH1:
				centercolumn = false;
				toggle_column ();
				break;
			//case Key.RIGHT:
			case ConsKey.ControlRight:
			case ConsKey.ControlRight1:
			case ConsKey.ControlRight2:
			case ConsKey.ControlRight3:
			case ConsKey.ControlL:
				centercolumn = true;
				toggle_column ();
				break;
			case ConsKey.ShiftRight:
			case 'L':
				if (centercolumn) {
					if (sel != null) {
						if (center.length () > 1) {
							center.remove (sel);
							panel.remove (sel);
							panel.prepend (sel);
							unowned List<Panel> tmp = panel.first ();
							if (tmp != null)
								selected = tmp;
							toggle_column ();
						}
					}
				}
				break;
			case ConsKey.ShiftLeft:
			case 'H':
				if (!centercolumn && sel != null) {
					center.remove (sel);
					panel.remove (sel);
					center.prepend (sel);
					unowned List<Panel> tmp = center.first ();
					if (tmp != null)
						selected = tmp;
					toggle_column ();
				}
				break;
			case ConsKey.ControlW:
				if (sel != null) {
					if (center.length () > 1) {
						center.remove (selected.data);
						panel.remove (selected.data);
						selected = center.first ();
					} else selected = null;
				}
				if (centercolumn) {
					if (selected == null) {
						selected = panel.first ();
						if (selected == null) {
							if (Dialog.do_yesno (this, "Do you want to quit?"))
								running = false;
							else selected = center.first ();
						} else {
							center.prepend (selected.data);
							panel.remove (selected.data);
							selected = center.first ();
						}
					}
				} else {
					selected = panel.first ();
					if (selected == null) {
						selected = center.first ();
						centercolumn = true;
					}
				}
				break;
			case ':':
				inputprompt = true;
				selected = null;
				break;
			case 'q':
				if (Dialog.do_yesno (this, "Do you want to quit?"))
					running = false;
				break;
			}
		}
		endwin ();
	}
}
