using Curses;

public class Panel {
	public Window win;
	public string title;
	public bool border;
	public int color;
	public int line;
	public int x;
	public int y;
	public int w;
	public int h;
	public Canvas canvas;

	public int scrollx;
	public int scrolly;

	public delegate bool ActivateCallback (int bar);

	public class Link {
		public int row;
		public ActivateCallback callback;

		public Link (int row, ActivateCallback cb) {
			this.row = row;
			this.callback = cb;
		}
	}

	protected int row;
	public List<Link?> links;

	public static int count = 0;
	public Panel (Canvas cv) {
		canvas = cv;
		move (LINES, COLS, 0, 0);
		win = new Window (LINES, COLS, 0, 0);
		lnk_init ();
		border = true;
		line = 3;
		color = 0;
		this.title = "panel %d".printf (++Panel.count);
	}

	public void move (int h, int w, int y, int x) {
		if (this.x != x || this.y != y || this.w != w || this.h != h) {
			win = new Window (h, w, y, x);
			win.bkgdset (COLOR_PAIR (color));
			this.x = x;
			this.y = y;
			this.w = w - 2;
			this.h = h - 2;
		}
	}

	/* body stuff */
	public virtual void render () {
		render_begin ();
		render_end ();
	}

	public void render_begin () {
		if (line<1) {
			scrolly += line-1;
			line = 1; 
		} else
		if (line>h) {
			scrolly += (line-h);
			line = h;
		}
		row = -scrolly + 2;
	}

	public int wline (string line) {
		row++;
		/* TODO: chop line if len>w */
		if (row>1 && row<=h)
			win.mvprintw (row, 2, line);
		return row;
	}

	public void render_end () {
		//win.clrtobot ();
		if (border) {
			win.mvprintw (1, 2, this.title);
			win.box (0, 0);
		} else {
			win.mvprintw (0, 0, this.title);
		}
		win.mvprintw (line, 1, ">"); // %d".printf (scrolly));
		//win.redrawwin ();
		win.refresh ();
	}

	public void scroll (int x, int y) {
		this.scrollx = x;
		this.scrolly = y;
	}

	public bool activate () {
		foreach (var lnk in links) {
			if (lnk.row == this.line)
				return lnk.callback (0);
		}
		return true;
	}

	/* XXX: This is hardly suboptimal */
	public void lnk_init () {
		links = new List<Link> ();
	}

	public void lnk_register (int row, ActivateCallback cb) {
		Link lnk = new Link (row, cb);
		links.append (lnk);
	}
}
